/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.apache.reef.runtime.standalone.client;

import org.apache.reef.runtime.local.LocalClasspathProvider;
import org.apache.reef.runtime.local.client.LocalJobSubmissionHandler;
import org.apache.reef.runtime.local.client.ExecutorServiceConstructor;
import org.apache.reef.client.parameters.DriverConfigurationProviders;
import org.apache.reef.runtime.common.client.CommonRuntimeConfiguration;
import org.apache.reef.runtime.common.client.DriverConfigurationProvider;
import org.apache.reef.runtime.common.client.api.JobSubmissionHandler;
import org.apache.reef.runtime.common.files.RuntimeClasspathProvider;
import org.apache.reef.runtime.standalone.client.parameters.NodeFolder;
import org.apache.reef.runtime.standalone.client.parameters.NodeListFilePath;
import org.apache.reef.runtime.standalone.client.parameters.RootFolder;
import org.apache.reef.runtime.standalone.client.parameters.SshPortNum;
import org.apache.reef.tang.ConfigurationProvider;
import org.apache.reef.tang.formats.*;

import java.util.concurrent.ExecutorService;

/**
 * A ConfigurationModule to configure the standalone resourcemanager.
 */
public final class StandaloneRuntimeConfiguration extends ConfigurationModuleBuilder {

  /**
   * The file containing a list of remote ssh nodes (i.e. `username@147.0.23.14`), separated by newlines.
   * The standalone runtime assumes the Driver to run on the same node as Client, thus a local file path is expected.
   * Currently, we expect that $JAVA_HOME is specified to the same directory in each of the nodes.
   */
  public static final RequiredParameter<String> NODE_LIST_FILE_PATH = new RequiredParameter<>();
  /**
   * Folder to save the shaded jar for the remote nodes.
   */
  public static final OptionalParameter<String> NODE_FOLDER = new OptionalParameter<>();

  /**
   * Folder to save the files generated by REEF. The corresponding sub-folders will be created for each node.
   */
  public static final OptionalParameter<String> RUNTIME_ROOT_FOLDER = new OptionalParameter<>();

  /**
   * The port number to attempt the ssh connection.
   */
  public static final OptionalParameter<Integer> SSH_PORT_NUM = new OptionalParameter<>();

  /**
   * Configuration provides whose Configuration will be merged into all Driver Configuration.
   */
  public static final OptionalImpl<ConfigurationProvider> DRIVER_CONFIGURATION_PROVIDERS = new OptionalImpl<>();

  /**
   * The ConfigurationModule for the standalone resourcemanager.
   */
  public static final ConfigurationModule CONF = new StandaloneRuntimeConfiguration()
      .merge(CommonRuntimeConfiguration.CONF)
          // Bind the standalone runtime
      .bindImplementation(JobSubmissionHandler.class, LocalJobSubmissionHandler.class)
      .bindImplementation(DriverConfigurationProvider.class, StandaloneDriverConfigurationProviderImpl.class)
      .bindConstructor(ExecutorService.class, ExecutorServiceConstructor.class)
      .bindImplementation(RuntimeClasspathProvider.class, LocalClasspathProvider.class)
          // Bind parameters of the standalone runtime
      .bindNamedParameter(NodeListFilePath.class, NODE_LIST_FILE_PATH)
      .bindNamedParameter(NodeFolder.class, NODE_FOLDER)
      .bindNamedParameter(RootFolder.class, RUNTIME_ROOT_FOLDER)
      .bindNamedParameter(SshPortNum.class, SSH_PORT_NUM)
      .bindSetEntry(DriverConfigurationProviders.class, DRIVER_CONFIGURATION_PROVIDERS)
      .build();


}
